"""This package contains tools to populate and validate the system"""
__author__ = """Bruno Rocha"""
__email__ = 'rochacbruno@gmail.com'
__version__ = '0.1.0'

from satellite_populate.decorators import populate_with  # noqa
from satellite_populate.main import load_data  # noqa
