"""To be implemented: a populator using CLI"""

# from satellite_populate.base import BasePopulator
#
#
# class CLIPopulator(BasePopulator):
#     """Populates system using hammer"""
#     # TO BE IMPLEMENTED
